#undef UNICODE

#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif

#include <Windows.h>
#include <WinSock2.h>
#include <WS2tcpip.h>

#pragma comment (lib, "Ws2_32.lib")
#pragma comment (lib, "Mswsock.lib")
#pragma comment (lib, "AdvApi32.lib")

#include <iostream>
#include <fstream>
#include <cassert>

#include "bb_client.h"

#define DEFAULT_PORT "27015"
static char data_buf[512];

// One single client socket
// When not open, this value is INVALID_SOCKET
static SOCKET connect_socket = INVALID_SOCKET;

// Stores various configuration settings
// Only update when a successful configuration occurs
static bb_settings current_config;
static bb_sweep_info current_sweep_info;

bool send_data(SOCKET my_socket, char *buf, int len, int flags = 0)
{
    int result = 0;
    while(len > 0) {
        result = send(my_socket, buf, len, flags);
        if(result > len) {
            printf("Error in send_data\n");
        }
        len -= result;
        buf += result;
    }

    return true;
}

bool recv_data(SOCKET my_socket, char *buf, int len, int flags = 0)
{
    int result = 0;
    while(len > 0) {
        result = recv(my_socket, buf, len, flags);
        if(result > len) {
            printf("Error in recv_data\n");
        }
        len -= result;
        buf += result;
    }

    return true;
}

// Can only be used if the connection was setup properly and
//  entirely
void end_socket_session()
{
    closesocket(connect_socket);
    connect_socket = INVALID_SOCKET;
    WSACleanup();
}

bbcStatus bbcConnect(const char *address)
{
    if(!address)
        return bbcNullPtrErr;

    // If session active return here
    if(connect_socket != INVALID_SOCKET) 
        return bbcConnectionAlreadyExistsErr;

    WSADATA wsa_data;
    addrinfo *result = NULL, *ptr = NULL, hints;
    connect_socket = INVALID_SOCKET;

    int wsa_result;

    // Initialize Winsock
    wsa_result = WSAStartup(MAKEWORD(2,2), &wsa_data);
    if(wsa_result != 0) {
        printf("WSAStartup failed with error: %d\n", wsa_result);
        return bbcSocketStartupErr;
    }

    memset(&hints, 0, sizeof(hints));
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;

    // Resolve the server address and port
    wsa_result = getaddrinfo(address, DEFAULT_PORT, &hints, &result);
    if(wsa_result != 0) {
        printf("getaddrinfo failed with error: %d\n", wsa_result);
        WSACleanup();
        return bbcSocketStartupErr;
    }

    // Attempt to connect to an address until one succeeds
    for(ptr = result; ptr != NULL; ptr = ptr->ai_next) {
        // Create a SOCKET for connecting to server
        connect_socket = socket(ptr->ai_family, ptr->ai_socktype, ptr->ai_protocol);
        if(connect_socket == INVALID_SOCKET) {
            printf("socket failed with error: %ld\n", WSAGetLastError());
            WSACleanup();
            return bbcSocketStartupErr;
        }

        // Connect to server.
        wsa_result = connect(connect_socket, ptr->ai_addr, (int)ptr->ai_addrlen);
        if(wsa_result == SOCKET_ERROR) {
            closesocket(connect_socket);
            connect_socket = INVALID_SOCKET;
            continue;
        }
        break;
    }

    freeaddrinfo(result);

    if(connect_socket == INVALID_SOCKET) {
        printf("Unable to connect to server!\n");
        WSACleanup();
        return bbcSocketStartupErr;
    }

    char ack;
    recv_data(connect_socket, &ack, 1, 0);
    if(ack != 'A') {
        end_socket_session();
        return bbcSocketStartupErr;
    }

    return bbcNoError;
}

bbcStatus bbcDisconnect()
{
    if(connect_socket == INVALID_SOCKET)
        return bbcNoError;

    char cmd = 'E';
    send_data(connect_socket, &cmd, 1, 0);

    end_socket_session();

    return bbcNoError;
}

// Attempt to remotely configure the device
// If successful, update sweep_info with new values
// If not successful, do not change the parameters
bbcStatus bbcConfigure(const bb_settings *settings, bb_sweep_info *sweep_info)
{
    if(!settings || !sweep_info)
        return bbcNullPtrErr;

    if(!connect_socket)
        return bbcNoConnectionErr;

    // C for configure
    data_buf[0] = 'C';
    memcpy(data_buf + 1, settings, sizeof(bb_settings));

    // Send the settings
    int result = send_data(connect_socket, data_buf, sizeof(bb_settings) + 1, 0);
    if(result == SOCKET_ERROR) {
        printf("send failed with error: %d\n", WSAGetLastError());
        end_socket_session();
        return bbcConnectionErr;
    }

    // Wait for ACK/NAK ('A'/'N'), sweep info attached
    result = recv_data(connect_socket, data_buf, sizeof(bb_sweep_info) + 1, 0);
    if(result > 0) {
        if(data_buf[0] == 'A') { // Successful config
            memcpy(&current_sweep_info, data_buf + 1, sizeof(bb_sweep_info));
            *sweep_info = current_sweep_info;
            current_config = *settings;
        } else if(data_buf[0] == 'N') { // Unsuccessful config
            return bbcInvalidConfigurationErr;
        } else { // Unknown
            assert(0);
        }
    }

    return bbcNoError;
}

bbcStatus bbcGetSweep(double *min, double *max)
{
    bbcStatus return_status = bbcNoError;

    if(!min || !max)
        return bbcNullPtrErr;

    if(!connect_socket)
        return bbcNoConnectionErr;

    // G for get sweep
    char msg = 'G';
    send_data(connect_socket, &msg, 1, 0);

    // Wait for ACK/NAK
    recv_data(connect_socket, &msg, 1, 0);

    if(msg == 'A' || msg == 'W') {
        recv_data(connect_socket, (char*)min, current_sweep_info.length * sizeof(double));
        recv_data(connect_socket, (char*)max, current_sweep_info.length * sizeof(double));

        if(msg == 'W') return_status = bbcIFOverload;

    } else {
        // NAK
        // Probably just haven't configured the device yet
        return_status = bbcInvalidConfigurationErr; 
    }

    return return_status;
}

bbcStatus bbcPreset()
{
    if(!connect_socket)
        return bbcNoConnectionErr;

    char msg = 'P';
    send_data(connect_socket, &msg, 1);
    recv_data(connect_socket, &msg, 1);

    if(msg != 'A') {
        end_socket_session();
        return bbcConnectionErr;
    }

    return bbcNoError;
}
