#undef UNICODE
#define WIN32_LEAN_AND_MEAN

#include <Windows.h>
#include <WinSock2.h>
#include <WS2tcpip.h>

#pragma comment(lib, "Ws2_32.lib")
#pragma comment(lib, "bb_api.lib")

#include <iostream>
#include <cassert>

#include "bb_api.h"

#define DEFAULT_PORT "27015"

// These must match between server/client
struct bb_settings {
    double center;
    double span;
    double ref_level;
    double rbw;
    double vbw;
    double sweep_time;
    int spur_reject;
    int detector;
};

struct bb_sweep_info {
    double start_freq;
    double bin_size;
    unsigned int length;
};

enum bbClientCommand {
    CommandConfigure, // 'C'
    CommandGetSweep,  // 'G'
    CommandPreset,    // 'P'
    CommandClose,     // 'E'
    CommandUnknown
};

bool send_data(SOCKET my_socket, char *buf, int len, int flags = 0)
{
    int result = 0;
    while(len > 0) {
        result = send(my_socket, buf, len, flags);
        if(result > len) {
            printf("Error in send_data\n");
        }
        len -= result;
        buf += result;
    }

    return true;
}

bool recv_data(SOCKET my_socket, char *buf, int len, int flags = 0)
{
    int result = 0;
    while(len > 0) {
        result = recv(my_socket, buf, len, flags);
        if(result > len) {
            printf("Error in recv_data\n");
        }
        len -= result;
        buf += result;
    }

    return true;
}

// Initialize the listening socket used for all client connections
// Return INVALID_SOCKET if failure
SOCKET get_listen_socket()
{
    WSADATA wsaData;
    SOCKET listen_socket = INVALID_SOCKET;
    struct addrinfo *result = NULL, hints;

    // Initialize Winsock
    int wsa_result = WSAStartup(MAKEWORD(2,2), &wsaData);
    if(wsa_result != 0) {
        printf("WSAStartup failed with error: %d\n", wsa_result);
        return INVALID_SOCKET;
    }

    ZeroMemory(&hints, sizeof(hints));
    hints.ai_family = AF_INET;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;
    hints.ai_flags = AI_PASSIVE;

    // Resolve the server address and port
    wsa_result = getaddrinfo(NULL, DEFAULT_PORT, &hints, &result);
    if(wsa_result != 0) {
        printf("getaddrinfo failed with error: %d\n", wsa_result);
        WSACleanup();
        return INVALID_SOCKET;
    }

    // Create a SOCKET for connecting to server
    listen_socket = socket(result->ai_family, result->ai_socktype, result->ai_protocol);
    if(listen_socket == INVALID_SOCKET) {
        printf("socket failed with error: %ld\n", WSAGetLastError());
        freeaddrinfo(result);
        WSACleanup();
        return INVALID_SOCKET;
    }

    // Setup the TCP listening socket
    wsa_result = bind(listen_socket, result->ai_addr, (int)result->ai_addrlen);
    if(wsa_result == SOCKET_ERROR) {
        printf("bind failed with error: %d\n", WSAGetLastError());
        freeaddrinfo(result);
        closesocket(listen_socket);
        WSACleanup();
        return INVALID_SOCKET;
    }

    freeaddrinfo(result);

    wsa_result = listen(listen_socket, SOMAXCONN);
    if (wsa_result == SOCKET_ERROR) {
        printf("listen failed with error: %d\n", WSAGetLastError());
        closesocket(listen_socket);
        WSACleanup();
        return INVALID_SOCKET;
    }

    return listen_socket;
}

// Get the single character 
bbClientCommand get_client_command(SOCKET client_socket)
{
    char command;
    int result = recv_data(client_socket, &command, 1, 0);

    switch(command) {
    case 'C':
        printf("Received configure command\n");
        return CommandConfigure;
        break;
    case 'G':
        printf("Received get sweep command\n");
        return CommandGetSweep;
        break;
    case 'P':
        printf("Recieved preset command\n");
        return CommandPreset;
        break;
    case 'E':
        printf("Received close command\n");
        return CommandClose;
        break;
    default:
        printf("Recieved unknown command\n");
        return CommandUnknown;
        break;
    }
}

void set_default_settings(bb_settings &settings)
{
    settings.center = 900.0e6;
    settings.span = 20.0e6;
    settings.rbw = 10.0e3;
    settings.vbw = 10.0e3;
    settings.ref_level = -20.0e3;
    settings.sweep_time = 0.001;
    settings.spur_reject = BB_NO_SPUR_REJECT;
    settings.detector = BB_MIN_AND_MAX;
}

bool configure_device(int id, bb_settings &settings, bb_sweep_info &info)
{
    bbConfigureAcquisition(id, settings.detector, BB_LOG_SCALE);
    bbConfigureCenterSpan(id, settings.center, settings.span);
    bbConfigureLevel(id, settings.ref_level, BB_AUTO_ATTEN);
    bbConfigureGain(id, BB_AUTO_GAIN);
    bbConfigureSweepCoupling(id, settings.rbw, settings.vbw, settings.sweep_time, 
        BB_NON_NATIVE_RBW, settings.spur_reject);
    bbConfigureProcUnits(id, BB_LOG);
    bbInitiate(id, BB_SWEEPING, 0);
    bbQueryTraceInfo(id, &info.length, &info.bin_size, &info.start_freq);
    
    return true;
}

bbStatus get_sweep(int id, int len, double *min, double *max)
{
    return bbFetchTrace(id, len, min, max);
}

// Try to connect a bb60
// 1) if can't connect bb60 close socket and return
// 2) otherwise accept commands until close command
// Do not close socket
void process_client_commands(SOCKET client_socket)
{
    bbClientCommand command = CommandUnknown;
    char cmd_buf[512];
    int result = 0;

    // Device information, can pass around, but lives here
    int dev_id = -1;
    bool device_configured = false;
    bb_settings device_settings, new_settings;
    bb_sweep_info device_sweep_info, new_info;

    set_default_settings(device_settings);

    // Attempt to open the device, the success of this
    //  changes whether the client connection stays open
    if(bbOpenDevice(&dev_id) != bbNoError) {
        send_data(client_socket, "N", 1, 0);
        return;
    } else {
        send_data(client_socket, "A", 1, 0);
    }

    // Prime the pump
    command = get_client_command(client_socket);

    // Infinite get command loop
    while(command != CommandUnknown && command != CommandClose) {
        switch(command) {
        case CommandConfigure:
            // Get new settings struct
            result = recv_data(client_socket, (char*)(&new_settings), sizeof(bb_settings), 0);
            // Try to configure with new struct
            if(configure_device(dev_id, new_settings, new_info)) {
                // Configured correctly prep and send ACK
                device_settings = new_settings;
                device_sweep_info = new_info;

                cmd_buf[0] = 'A';
                memcpy(cmd_buf + 1, &device_sweep_info, sizeof(bb_sweep_info));
                send_data(client_socket, cmd_buf, sizeof(bb_sweep_info) + 1, 0);

                device_configured = true;
            } else {
                // Did not configure correctly
                // Configure with old settings again, and send NAK
                configure_device(dev_id, device_settings, device_sweep_info);

                cmd_buf[0] = 'N';
                memcpy(cmd_buf + 1, &device_sweep_info, sizeof(bb_sweep_info));
                send_data(client_socket, cmd_buf, sizeof(bb_sweep_info) + 1, 0);
            }
            break;
        case CommandGetSweep:
            // Device must be configured once before get sweep 
            if(!device_configured) {
                send_data(client_socket, "N", 1);
                break;
            } else {
                // Device is configured
                double *min = new double[device_sweep_info.length],
                    *max = new double[device_sweep_info.length];
                bbStatus sweep_status = get_sweep(dev_id, device_sweep_info.length, min, max);
                if(sweep_status == bbNoError) {
                    send_data(client_socket, "A", 1);
                    send_data(client_socket, (char*)min, 
                        sizeof(double) * device_sweep_info.length);
                    send_data(client_socket, (char*)max, 
                        sizeof(double) * device_sweep_info.length);
                } else if(sweep_status == bbADCOverflow) {
                    send_data(client_socket, "W", 1); // 'W' indicates overflow
                    send_data(client_socket, (char*)min, 
                        sizeof(double) * device_sweep_info.length);
                    send_data(client_socket, (char*)max, 
                        sizeof(double) * device_sweep_info.length);
                } else {
                    send_data(client_socket, "N", 1);
                }
                delete [] min;
                delete [] max;
            }
            break;
        case CommandPreset:
            bbPreset(dev_id);
            bbCloseDevice(dev_id);
            // Three second wait is almost always appropriate
            Sleep(3000);
            // If device fails to open again, then return as device 
            //  is already closed, nothing else to do
            if(bbOpenDevice(&dev_id) != bbNoError) {
                send_data(client_socket, "N", 1);
                dev_id = -1;
                return;
            } else {
                configure_device(dev_id, device_settings, device_sweep_info);
                send_data(client_socket, "A", 1);
            }
            break;
        default:
            printf("Invalid command\n");
            assert(0);
            break;
        }
        // Pre-fetch next command for while() loop
        command = get_client_command(client_socket);
    }

    bbCloseDevice(dev_id);
}

int main(int argc, char **argv)
{
    SOCKET listen_socket = get_listen_socket();

    if(listen_socket == INVALID_SOCKET) {
        printf("Unable to create socket, exiting program\n");
        return 1;
    }

    std::cout << "Listening on port " << DEFAULT_PORT << std::endl;

    SOCKET client_socket = INVALID_SOCKET;

    // Loop forever
    // 1) Connect with new client
    // 2) Process commands until close
    // 3) Repeat
    do {
        // 1 : Accept new client connection
        client_socket = accept(listen_socket, NULL, NULL);
        if(client_socket == INVALID_SOCKET) {
            printf("accept failed with error: %d\n", WSAGetLastError());
            closesocket(listen_socket);
            WSACleanup();
            return 1;
        }
        // 2 : Process commands for client
        printf("Client connected\n");
        process_client_commands(client_socket);

        closesocket(client_socket);
        client_socket = INVALID_SOCKET;
    } while(true);

    return 0;
}